package ru.dbotthepony.kstarbound.math

import java.nio.ByteBuffer
import java.nio.ByteOrder
import java.nio.FloatBuffer
import kotlin.math.PI
import kotlin.math.tan

interface IMatrixLike {
	val columns: Int
	val rows: Int
}

interface IMatrixLikeGetterI {
	operator fun get(row: Int, column: Int): Int
}

interface IMatrixLikeFloat : IMatrixLike {
	operator fun get(row: Int, column: Int): Float
}

interface IMatrix : IMatrixLike {
	operator fun plus(other: IMatrix): IMatrix
	operator fun minus(other: IMatrix): IMatrix
	operator fun times(other: IMatrix): IMatrix
}

interface FloatMatrix<T : FloatMatrix<T>> : IMatrix, IMatrixLikeFloat {
	operator fun plus(other: Float): T
	operator fun minus(other: Float): T
	operator fun times(other: Float): T
	operator fun div(other: Float): T

	override operator fun plus(other: IMatrix): T
	override operator fun minus(other: IMatrix): T
	override operator fun times(other: IMatrix): T

	/**
	 * Если матрица больше или меньше, считать что всевозможные остальные координаты равны единице (не менять)
	 */
	fun scale(x: Float, y: Float = 1f, z: Float = 1f, w: Float = 1f): T

	fun scale(value: Vector4f) = scale(value.x, value.y, value.z, value.w)

	fun translate(x: Float = 0f, y: Float = 0f, z: Float = 0f): T
	fun translate(vector3f: Vector3f) = translate(vector3f.x, vector3f.y, vector3f.z)

	fun translateWithScale(x: Float = 0f, y: Float = 0f, z: Float = 0f): T
	fun translateWithScale(vector3f: Vector3f) = translateWithScale(vector3f.x, vector3f.y, vector3f.z)

	/**
	 * Выдает массив в готовом для OpenGL виде (строка -> столбец) по умолчанию
	 */
	fun toFloatArray(columnMajor: Boolean = true): FloatArray {
		val buff = FloatArray(rows * columns)

		if (columnMajor) {
			for (row in 0 until rows) {
				for (column in 0 until columns) {
					buff[row + rows * column] = this[row, column]
				}
			}
		} else {
			for (row in 0 until rows) {
				for (column in 0 until columns) {
					buff[row * columns + column] = this[row, column]
				}
			}
		}

		return buff
	}
}

interface MutableFloatMatrix<T : MutableFloatMatrix<T>> : FloatMatrix<T> {
	operator fun set(row: Int, column: Int, value: Float)
}

abstract class AbstractMatrix4f<T : AbstractMatrix4f<T>> : FloatMatrix<T> {
	companion object {
		private val directBuffer by lazy { ByteBuffer.allocateDirect(4 * 4 * 4).also { it.order(ByteOrder.nativeOrder()) }.asFloatBuffer()!! }
	}

	/**
	 * Наполняет синглтон и возвращает DirectFloatBuffer который необходимо использовать сразу,
	 * нет никаких гарантий что он не будет изменён в дальнейшем
	 *
	 * Функция предназначена исключительно для загрузки данной матрицы в память видеокарты
	 *
	 * Данный буфер всегда будет одним и тем же, но его содержимое обновлено
	 */
	fun toFloatBuffer(columnMajor: Boolean = true): FloatBuffer {
		val directBuffer = directBuffer
		directBuffer.position(0)

		if (columnMajor) {
			directBuffer.put(m00)
			directBuffer.put(m10)
			directBuffer.put(m20)
			directBuffer.put(m30)

			directBuffer.put(m01)
			directBuffer.put(m11)
			directBuffer.put(m21)
			directBuffer.put(m31)

			directBuffer.put(m02)
			directBuffer.put(m12)
			directBuffer.put(m22)
			directBuffer.put(m32)

			directBuffer.put(m03)
			directBuffer.put(m13)
			directBuffer.put(m23)
			directBuffer.put(m33)
		} else {
			directBuffer.put(m00)
			directBuffer.put(m01)
			directBuffer.put(m02)
			directBuffer.put(m03)

			directBuffer.put(m10)
			directBuffer.put(m11)
			directBuffer.put(m12)
			directBuffer.put(m13)

			directBuffer.put(m20)
			directBuffer.put(m21)
			directBuffer.put(m22)
			directBuffer.put(m23)

			directBuffer.put(m30)
			directBuffer.put(m31)
			directBuffer.put(m32)
			directBuffer.put(m33)
		}

		directBuffer.position(0)
		return directBuffer
	}

	abstract val m00: Float; abstract val m01: Float; abstract val m02: Float; abstract val m03: Float
	abstract val m10: Float; abstract val m11: Float; abstract val m12: Float; abstract val m13: Float
	abstract val m20: Float; abstract val m21: Float; abstract val m22: Float; abstract val m23: Float
	abstract val m30: Float; abstract val m31: Float; abstract val m32: Float; abstract val m33: Float

	override val columns: Int
		get() = 4

	override val rows: Int
		get() = 4

	override fun get(row: Int, column: Int): Float {
		return when (column) {
			0 -> when (row) {
				0 -> m00
				1 -> m10
				2 -> m20
				3 -> m30
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01
				1 -> m11
				2 -> m21
				3 -> m31
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			2 -> when (row) {
				0 -> m02
				1 -> m12
				2 -> m22
				3 -> m32
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			3 -> when (row) {
				0 -> m03
				1 -> m13
				2 -> m23
				3 -> m33
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}

	protected abstract fun createOrModify(
		m00: Float, m01: Float, m02: Float, m03: Float,
		m10: Float, m11: Float, m12: Float, m13: Float,
		m20: Float, m21: Float, m22: Float, m23: Float,
		m30: Float, m31: Float, m32: Float, m33: Float,
	): T

	override fun plus(other: IMatrix): T {
		if (other !is FloatMatrix<*>) {
			throw IllegalArgumentException("Can not use $other for addition")
		}

		if (other.columns != 4 || other.rows != 4) {
			throw IllegalArgumentException("Concrete Matrix4f can only use 4x4 matrixes")
		}

		val m00: Float; val m01: Float; val m02: Float; val m03: Float;
		val m10: Float; val m11: Float; val m12: Float; val m13: Float;
		val m20: Float; val m21: Float; val m22: Float; val m23: Float;
		val m30: Float; val m31: Float; val m32: Float; val m33: Float;

		if (other is Matrix4f) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02; m03 = other.m03;
			m10 = other.m10; m11 = other.m11; m12 = other.m12; m13 = other.m13;
			m20 = other.m20; m21 = other.m21; m22 = other.m22; m23 = other.m23;
			m30 = other.m30; m31 = other.m31; m32 = other.m32; m33 = other.m33;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2]; m03 = other[0, 3];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2]; m13 = other[1, 3];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2]; m23 = other[2, 3];
			m30 = other[3, 0]; m31 = other[3, 1]; m32 = other[3, 2]; m33 = other[3, 3];
		}

		return createOrModify(
			this.m00 + m00, this.m01 + m01, this.m02 + m02, this.m03 + m03,
			this.m10 + m10, this.m11 + m11, this.m12 + m12, this.m13 + m13,
			this.m20 + m20, this.m21 + m21, this.m22 + m22, this.m23 + m23,
			this.m30 + m30, this.m31 + m31, this.m32 + m32, this.m33 + m33,
		)
	}

	override fun minus(other: IMatrix): T {
		if (other !is FloatMatrix<*>) {
			throw IllegalArgumentException("Can not use $other for subtraction")
		}

		if (other.columns != 4 || other.rows != 4) {
			throw IllegalArgumentException("Concrete Matrix4f can only use 4x4 matrices")
		}

		val m00: Float; val m01: Float; val m02: Float; val m03: Float;
		val m10: Float; val m11: Float; val m12: Float; val m13: Float;
		val m20: Float; val m21: Float; val m22: Float; val m23: Float;
		val m30: Float; val m31: Float; val m32: Float; val m33: Float;

		if (other is Matrix4f) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02; m03 = other.m03;
			m10 = other.m10; m11 = other.m11; m12 = other.m12; m13 = other.m13;
			m20 = other.m20; m21 = other.m21; m22 = other.m22; m23 = other.m23;
			m30 = other.m30; m31 = other.m31; m32 = other.m32; m33 = other.m33;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2]; m03 = other[0, 3];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2]; m13 = other[1, 3];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2]; m23 = other[2, 3];
			m30 = other[3, 0]; m31 = other[3, 1]; m32 = other[3, 2]; m33 = other[3, 3];
		}

		return createOrModify(
			this.m00 - m00, this.m01 - m01, this.m02 - m02, this.m03 - m03,
			this.m10 - m10, this.m11 - m11, this.m12 - m12, this.m13 - m13,
			this.m20 - m20, this.m21 - m21, this.m22 - m22, this.m23 - m23,
			this.m30 - m30, this.m31 - m31, this.m32 - m32, this.m33 - m33,
		)
	}

	override fun plus(other: Float): T {
		return createOrModify(
			this.m00 + other, this.m01 + other, this.m02 + other, this.m03 + other,
			this.m10 + other, this.m11 + other, this.m12 + other, this.m13 + other,
			this.m20 + other, this.m21 + other, this.m22 + other, this.m23 + other,
			this.m30 + other, this.m31 + other, this.m32 + other, this.m33 + other,
		)
	}

	override fun minus(other: Float): T {
		return createOrModify(
			this.m00 - other, this.m01 - other, this.m02 - other, this.m03 - other,
			this.m10 - other, this.m11 - other, this.m12 - other, this.m13 - other,
			this.m20 - other, this.m21 - other, this.m22 - other, this.m23 - other,
			this.m30 - other, this.m31 - other, this.m32 - other, this.m33 - other,
		)
	}

	override fun times(other: Float): T {
		return createOrModify(
			this.m00 * other, this.m01 * other, this.m02 * other, this.m03 * other,
			this.m10 * other, this.m11 * other, this.m12 * other, this.m13 * other,
			this.m20 * other, this.m21 * other, this.m22 * other, this.m23 * other,
			this.m30 * other, this.m31 * other, this.m32 * other, this.m33 * other,
		)
	}

	override fun scale(x: Float, y: Float, z: Float, w: Float): T {
		return createOrModify(
			this.m00 * x, this.m01, this.m02, this.m03,
			this.m10, this.m11 * y, this.m12, this.m13,
			this.m20, this.m21, this.m22 * z, this.m23,
			this.m30, this.m31, this.m32, this.m33 * w,
		)
	}

	override fun translate(x: Float, y: Float, z: Float): T {
		return createOrModify(
			this.m00, this.m01, this.m02, this.m03 + x,
			this.m10, this.m11, this.m12, this.m13 + y,
			this.m20, this.m21, this.m22, this.m23 + z,
			this.m30, this.m31, this.m32, this.m33,
		)
	}

	open val translation: Vector3f
		get() {
			return Vector3f(m03, m13, m23)
		}

	override fun div(other: Float): T {
		return createOrModify(
			this.m00 / other, this.m01 / other, this.m02 / other, this.m03 / other,
			this.m10 / other, this.m11 / other, this.m12 / other, this.m13 / other,
			this.m20 / other, this.m21 / other, this.m22 / other, this.m23 / other,
			this.m30 / other, this.m31 / other, this.m32 / other, this.m33 / other,
		)
	}

	override fun times(other: IMatrix): T {
		if (other !is FloatMatrix<*>) {
			throw IllegalArgumentException("Can not use $other for multiplication")
		}

		if (other.columns != 4 || other.rows != 4) {
			throw IllegalArgumentException("Concrete Matrix4f can only use 4x4 matrices")
		}

		val m00: Float; val m01: Float; val m02: Float; val m03: Float;
		val m10: Float; val m11: Float; val m12: Float; val m13: Float;
		val m20: Float; val m21: Float; val m22: Float; val m23: Float;
		val m30: Float; val m31: Float; val m32: Float; val m33: Float;

		if (other is Matrix4f) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02; m03 = other.m03;
			m10 = other.m10; m11 = other.m11; m12 = other.m12; m13 = other.m13;
			m20 = other.m20; m21 = other.m21; m22 = other.m22; m23 = other.m23;
			m30 = other.m30; m31 = other.m31; m32 = other.m32; m33 = other.m33;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2]; m03 = other[0, 3];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2]; m13 = other[1, 3];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2]; m23 = other[2, 3];
			m30 = other[3, 0]; m31 = other[3, 1]; m32 = other[3, 2]; m33 = other[3, 3];
		}

		// первый столбец
		val newm00 =
			this.m00 * m00 +
					this.m01 * m10 +
					this.m02 * m20 +
					this.m03 * m30

		val newm10 =
			this.m10 * m00 +
					this.m11 * m10 +
					this.m12 * m20 +
					this.m13 * m30

		val newm20 =
			this.m20 * m00 +
					this.m21 * m10 +
					this.m22 * m20 +
					this.m23 * m30

		val newm30 =
			this.m30 * m00 +
					this.m31 * m10 +
					this.m32 * m20 +
					this.m33 * m30

		// второй столбец
		val newm01 =
			this.m00 * m01 +
					this.m01 * m11 +
					this.m02 * m21 +
					this.m03 * m31

		val newm11 =
			this.m10 * m01 +
					this.m11 * m11 +
					this.m12 * m21 +
					this.m13 * m31

		val newm21 =
			this.m20 * m01 +
					this.m21 * m11 +
					this.m22 * m21 +
					this.m23 * m31

		val newm31 =
			this.m30 * m01 +
					this.m31 * m11 +
					this.m32 * m21 +
					this.m33 * m31

		// третий столбец
		val newm02 =
			this.m00 * m02 +
					this.m01 * m12 +
					this.m02 * m22 +
					this.m03 * m32

		val newm12 =
			this.m10 * m02 +
					this.m11 * m12 +
					this.m12 * m22 +
					this.m13 * m32

		val newm22 =
			this.m20 * m02 +
					this.m21 * m12 +
					this.m22 * m22 +
					this.m23 * m32

		val newm32 =
			this.m30 * m02 +
					this.m31 * m12 +
					this.m32 * m22 +
					this.m33 * m32

		// четвёртый столбец
		val newm03 =
			this.m00 * m03 +
					this.m01 * m13 +
					this.m02 * m23 +
					this.m03 * m33

		val newm13 =
			this.m10 * m03 +
					this.m11 * m13 +
					this.m12 * m23 +
					this.m13 * m33

		val newm23 =
			this.m20 * m03 +
					this.m21 * m13 +
					this.m22 * m23 +
					this.m23 * m33

		val newm33 =
			this.m30 * m03 +
					this.m31 * m13 +
					this.m32 * m23 +
					this.m33 * m33

		return createOrModify(
			newm00, newm01, newm02, newm03,
			newm10, newm11, newm12, newm13,
			newm20, newm21, newm22, newm23,
			newm30, newm31, newm32, newm33,
		)
	}

	override fun translateWithScale(x: Float, y: Float, z: Float): T {
		return createOrModify(
			m00, m01, m02, m03 + x * m00 + y * m01 + z * m02,
			m10, m11, m12, m13 + x * m10 + y * m11 + z * m12,
			m20, m21, m22, m23 + x * m20 + y * m21 + z * m22,
			m30, m31, m32, m33,
		)
	}
}

data class Matrix4f(
	override val m00: Float = 1f, override val m01: Float = 0f, override val m02: Float = 0f, override val m03: Float = 0f,
	override val m10: Float = 0f, override val m11: Float = 1f, override val m12: Float = 0f, override val m13: Float = 0f,
	override val m20: Float = 0f, override val m21: Float = 0f, override val m22: Float = 1f, override val m23: Float = 0f,
	override val m30: Float = 0f, override val m31: Float = 0f, override val m32: Float = 0f, override val m33: Float = 1f,
) : AbstractMatrix4f<Matrix4f>() {
	override fun createOrModify(
		m00: Float, m01: Float, m02: Float, m03: Float,
		m10: Float, m11: Float, m12: Float, m13: Float,
		m20: Float, m21: Float, m22: Float, m23: Float,
		m30: Float, m31: Float, m32: Float, m33: Float,
	): Matrix4f {
		return Matrix4f(
			m00 = m00, m01 = m01, m02 = m02, m03 = m03,
			m10 = m10, m11 = m11, m12 = m12, m13 = m13,
			m20 = m20, m21 = m21, m22 = m22, m23 = m23,
			m30 = m30, m31 = m31, m32 = m32, m33 = m33,
		)
	}

	fun toMutableMatrix(): MutableMatrix4f {
		return MutableMatrix4f(
			m00 = m00, m01 = m01, m02 = m02, m03 = m03,
			m10 = m10, m11 = m11, m12 = m12, m13 = m13,
			m20 = m20, m21 = m21, m22 = m22, m23 = m23,
			m30 = m30, m31 = m31, m32 = m32, m33 = m33,
		)
	}

	companion object {
		val IDENTITY = Matrix4f()

		val SCREEN_FLIP = IDENTITY.let {
			return@let it * Vector3f.FORWARD.rotateAroundThis(-PI / 2)
		}

		/**
		 * Возвращает матрицу ортографической проекции, с ИНВЕНТИРОВАННОЙ y координатой, и с добавлением 2f
		 *
		 * Т.е. любое значение компоненты вектора y будет иметь противоположный знак после перемножения на данную матрицу
		 *
		 * Смысл данного изменения знака в преобразовании экранных координат OpenGL к вменяемому виду. Многие примеры указывают Z как отрицательную компоненту,
		 * что так же "убирает" это недоумение, только вот у нас Z это глубина, а не высота
		 *
		 * y = 0 будет соответствовать верхнему левому углу окна
		 */
		fun ortho(left: Float, right: Float, bottom: Float, top: Float, zNear: Float, zFar: Float): Matrix4f {
			return Matrix4f(
				m00 = 2f / (right - left),
				m11 = -2f / (top - bottom),
				m22 = 2f / (zFar - zNear),
				m03 = -(right + left) / (right - left),
				m13 = -(top + bottom) / (top - bottom) + 2f,
				m23 = -(zFar + zNear) / (zFar - zNear)
			)
		}

		/**
		 * Возвращает матрицу ортографической проекции, без инвентирования
		 *
		 * y = 0 будет соответствовать нижнему левому углу окна
		 */
		fun orthoDirect(left: Float, right: Float, bottom: Float, top: Float, zNear: Float, zFar: Float): Matrix4f {
			return Matrix4f(
				m00 = 2f / (right - left),
				m11 = 2f / (top - bottom),
				m22 = 2f / (zFar - zNear),
				m03 = -(right + left) / (right - left),
				m13 = -(top + bottom) / (top - bottom),
				m23 = -(zFar + zNear) / (zFar - zNear)
			)
		}

		fun perspective(fov: Float, zFar: Float, zNear: Float): Matrix4f {
			val scale = (1.0 / (tan(Math.toRadians(fov.toDouble()) / 2.0))).toFloat()
			val r = zFar - zNear

			return Matrix4f(
				m00 = scale,
				m11 = scale,
				m22 = -zFar / r,
				m23 = -1f,
				m32 = -zFar * zNear / r,
			)
		}
	}
}

data class MutableMatrix4f(
	override var m00: Float = 1f, override var m01: Float = 0f, override var m02: Float = 0f, override var m03: Float = 0f,
	override var m10: Float = 0f, override var m11: Float = 1f, override var m12: Float = 0f, override var m13: Float = 0f,
	override var m20: Float = 0f, override var m21: Float = 0f, override var m22: Float = 1f, override var m23: Float = 0f,
	override var m30: Float = 0f, override var m31: Float = 0f, override var m32: Float = 0f, override var m33: Float = 1f,
) : AbstractMatrix4f<MutableMatrix4f>() {
	override fun createOrModify(
		m00: Float, m01: Float, m02: Float, m03: Float,
		m10: Float, m11: Float, m12: Float, m13: Float,
		m20: Float, m21: Float, m22: Float, m23: Float,
		m30: Float, m31: Float, m32: Float, m33: Float,
	): MutableMatrix4f {
		this.m00 = m00; this.m01 = m01; this.m02 = m02; this.m03 = m03
		this.m10 = m10; this.m11 = m11; this.m12 = m12; this.m13 = m13
		this.m20 = m20; this.m21 = m21; this.m22 = m22; this.m23 = m23
		this.m30 = m30; this.m31 = m31; this.m32 = m32; this.m33 = m33

		return this
	}

	override var translation: Vector3f
		get() {
			return Vector3f(m03, m13, m23)
		}
		set(value) {
			translate(value)
		}

	fun set(row: Int, column: Int, value: Float) {
		when (column) {
			0 -> when (row) {
				0 -> m00 = value
				1 -> m10 = value
				2 -> m20 = value
				3 -> m30 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01 = value
				1 -> m11 = value
				2 -> m21 = value
				3 -> m31 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			2 -> when (row) {
				0 -> m02 = value
				1 -> m12 = value
				2 -> m22 = value
				3 -> m32 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			3 -> when (row) {
				0 -> m03 = value
				1 -> m13 = value
				2 -> m23 = value
				3 -> m33 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}
}

abstract class AbstractMatrix3f<T : AbstractMatrix3f<T>> : FloatMatrix<T> {
	companion object {
		private val directBuffer by lazy { ByteBuffer.allocateDirect(4 * 3 * 3).also { it.order(ByteOrder.nativeOrder()) }.asFloatBuffer()!! }
	}

	/**
	 * Наполняет синглтон и возвращает DirectFloatBuffer который необходимо использовать сразу,
	 * нет никаких гарантий что он не будет изменён в дальнейшем
	 *
	 * Функция предназначена исключительно для загрузки данной матрицы в память видеокарты
	 *
	 * Данный буфер всегда будет одним и тем же, но его содержимое обновлено
	 */
	fun toFloatBuffer(columnMajor: Boolean = true): FloatBuffer {
		val directBuffer = directBuffer
		directBuffer.position(0)

		if (columnMajor) {
			directBuffer.put(m00)
			directBuffer.put(m10)
			directBuffer.put(m20)

			directBuffer.put(m01)
			directBuffer.put(m11)
			directBuffer.put(m21)

			directBuffer.put(m02)
			directBuffer.put(m12)
			directBuffer.put(m22)
		} else {
			directBuffer.put(m00)
			directBuffer.put(m01)
			directBuffer.put(m02)

			directBuffer.put(m10)
			directBuffer.put(m11)
			directBuffer.put(m12)

			directBuffer.put(m20)
			directBuffer.put(m21)
			directBuffer.put(m22)
		}

		directBuffer.position(0)
		return directBuffer
	}

	abstract val m00: Float; abstract val m01: Float; abstract val m02: Float
	abstract val m10: Float; abstract val m11: Float; abstract val m12: Float
	abstract val m20: Float; abstract val m21: Float; abstract val m22: Float

	override val columns: Int
		get() = 3

	override val rows: Int
		get() = 3

	override fun get(row: Int, column: Int): Float {
		return when (column) {
			0 -> when (row) {
				0 -> m00
				1 -> m10
				2 -> m20
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01
				1 -> m11
				2 -> m21
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			2 -> when (row) {
				0 -> m02
				1 -> m12
				2 -> m22
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}

	protected abstract fun createOrModify(
		m00: Float, m01: Float, m02: Float,
		m10: Float, m11: Float, m12: Float,
		m20: Float, m21: Float, m22: Float,
	): T

	override fun plus(other: IMatrix): T {
		if (other !is FloatMatrix<*>) {
			throw IllegalArgumentException("Can not use $other for addition")
		}

		if (other.columns != 3 || other.rows != 3) {
			throw IllegalArgumentException("Concrete Matrix3f can only use 3x3 matrices")
		}

		val m00: Float; val m01: Float; val m02: Float;
		val m10: Float; val m11: Float; val m12: Float;
		val m20: Float; val m21: Float; val m22: Float;

		if (other is Matrix3f) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02;
			m10 = other.m10; m11 = other.m11; m12 = other.m12;
			m20 = other.m20; m21 = other.m21; m22 = other.m22;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2];
		}

		return createOrModify(
			this.m00 + m00, this.m01 + m01, this.m02 + m02,
			this.m10 + m10, this.m11 + m11, this.m12 + m12,
			this.m20 + m20, this.m21 + m21, this.m22 + m22,
		)
	}

	override fun minus(other: IMatrix): T {
		if (other !is FloatMatrix<*>) {
			throw IllegalArgumentException("Can not use $other for subtraction")
		}

		if (other.columns != 3 || other.rows != 3) {
			throw IllegalArgumentException("Concrete Matrix3f can only use 3x3 matrices")
		}

		val m00: Float; val m01: Float; val m02: Float;
		val m10: Float; val m11: Float; val m12: Float;
		val m20: Float; val m21: Float; val m22: Float;

		if (other is Matrix3f) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02;
			m10 = other.m10; m11 = other.m11; m12 = other.m12;
			m20 = other.m20; m21 = other.m21; m22 = other.m22;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2];
		}

		return createOrModify(
			this.m00 - m00, this.m01 - m01, this.m02 - m02,
			this.m10 - m10, this.m11 - m11, this.m12 - m12,
			this.m20 - m20, this.m21 - m21, this.m22 - m22,
		)
	}

	override fun plus(other: Float): T {
		return createOrModify(
			this.m00 + other, this.m01 + other, this.m02 + other,
			this.m10 + other, this.m11 + other, this.m12 + other,
			this.m20 + other, this.m21 + other, this.m22 + other,
		)
	}

	override fun minus(other: Float): T {
		return createOrModify(
			this.m00 - other, this.m01 - other, this.m02 - other,
			this.m10 - other, this.m11 - other, this.m12 - other,
			this.m20 - other, this.m21 - other, this.m22 - other,
		)
	}

	override fun times(other: Float): T {
		return createOrModify(
			this.m00 * other, this.m01 * other, this.m02 * other,
			this.m10 * other, this.m11 * other, this.m12 * other,
			this.m20 * other, this.m21 * other, this.m22 * other,
		)
	}

	override fun div(other: Float): T {
		return createOrModify(
			this.m00 / other, this.m01 / other, this.m02 / other,
			this.m10 / other, this.m11 / other, this.m12 / other,
			this.m20 / other, this.m21 / other, this.m22 / other,
		)
	}

	override fun scale(x: Float, y: Float, z: Float, w: Float): T {
		return createOrModify(
			this.m00 * x, this.m01, this.m02,
			this.m10, this.m11 * y, this.m12,
			this.m20, this.m21, this.m22 * z,
		)
	}

	override fun translate(x: Float, y: Float, z: Float): T {
		return createOrModify(
			this.m00, this.m01, this.m02 + x,
			this.m10, this.m11, this.m12 + y,
			this.m20, this.m21, this.m22,
		)
	}

	override fun times(other: IMatrix): T {
		if (other !is FloatMatrix<*>) {
			throw IllegalArgumentException("Can not use $other for multiplication")
		}

		if (other.columns != 3 || other.rows != 3) {
			throw IllegalArgumentException("Concrete Matrix3f can only use 3x3 matrixes")
		}

		val m00: Float; val m01: Float; val m02: Float;
		val m10: Float; val m11: Float; val m12: Float;
		val m20: Float; val m21: Float; val m22: Float;

		if (other is Matrix3f) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02;
			m10 = other.m10; m11 = other.m11; m12 = other.m12;
			m20 = other.m20; m21 = other.m21; m22 = other.m22;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2];
		}

		// первый столбец
		val newm00 =
			this.m00 * m00 +
					this.m01 * m10 +
					this.m02 * m20

		val newm10 =
			this.m10 * m00 +
					this.m11 * m10 +
					this.m12 * m20

		val newm20 =
			this.m20 * m00 +
					this.m21 * m10 +
					this.m22 * m20

		// второй столбец
		val newm01 =
			this.m00 * m01 +
					this.m01 * m11 +
					this.m02 * m21

		val newm11 =
			this.m10 * m01 +
					this.m11 * m11 +
					this.m12 * m21

		val newm21 =
			this.m20 * m01 +
					this.m21 * m11 +
					this.m22 * m21

		// третий столбец
		val newm02 =
			this.m00 * m02 +
					this.m01 * m12 +
					this.m02 * m22

		val newm12 =
			this.m10 * m02 +
					this.m11 * m12 +
					this.m12 * m22

		val newm22 =
			this.m20 * m02 +
					this.m21 * m12 +
					this.m22 * m22

		return createOrModify(
			newm00, newm01, newm02,
			newm10, newm11, newm12,
			newm20, newm21, newm22,
		)
	}

	override fun translateWithScale(x: Float, y: Float, z: Float): T {
		return createOrModify(
			m00, m01, m02 + x * m00 + y * m01,
			m10, m11, m12 + x * m10 + y * m11,
			m20, m21, m22,
		)
	}
}

data class Matrix3f(
	override val m00: Float = 1f, override val m01: Float = 0f, override val m02: Float = 0f,
	override val m10: Float = 0f, override val m11: Float = 1f, override val m12: Float = 0f,
	override val m20: Float = 0f, override val m21: Float = 0f, override val m22: Float = 1f,
) : AbstractMatrix3f<Matrix3f>() {
	override fun createOrModify(
		m00: Float, m01: Float, m02: Float,
		m10: Float, m11: Float, m12: Float,
		m20: Float, m21: Float, m22: Float,
	): Matrix3f {
		return Matrix3f(
			m00 = m00, m01 = m01, m02 = m02,
			m10 = m10, m11 = m11, m12 = m12,
			m20 = m20, m21 = m21, m22 = m22,
		)
	}

	companion object {
		val IDENTITY = Matrix3f()
	}
}

data class MutableMatrix3f(
	override var m00: Float = 1f, override var m01: Float = 0f, override var m02: Float = 0f,
	override var m10: Float = 0f, override var m11: Float = 1f, override var m12: Float = 0f,
	override var m20: Float = 0f, override var m21: Float = 0f, override var m22: Float = 1f,
) : AbstractMatrix3f<MutableMatrix3f>() {
	override fun createOrModify(
		m00: Float, m01: Float, m02: Float,
		m10: Float, m11: Float, m12: Float,
		m20: Float, m21: Float, m22: Float,
	): MutableMatrix3f {
		this.m00 = m00; this.m01 = m01; this.m02 = m02
		this.m10 = m10; this.m11 = m11; this.m12 = m12
		this.m20 = m20; this.m21 = m21; this.m22 = m22

		return this
	}

	operator fun set(row: Int, column: Int, value: Float) {
		when (column) {
			0 -> when (row) {
				0 -> m00 = value
				1 -> m10 = value
				2 -> m20 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01 = value
				1 -> m11 = value
				2 -> m21 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			2 -> when (row) {
				0 -> m02 = value
				1 -> m12 = value
				2 -> m22 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}
}
